# Copyright 2023 Niklas Birth
#
# This file is part of CoCoPyE.
#
# CoCoPyE is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# CoCoPyE is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with CoCoPyE. If not, see <https://www.gnu.org/licenses/>.

import os
import shutil
import subprocess
import sys
import importlib.util
import tempfile

import numpy as np
import pandas as pd
import pkg_resources

from appdirs import user_data_dir, user_config_dir
from numba import set_num_threads

from .. import config
from ..external import check_and_download_dependencies
from ..external.data import update_cocopye_db
from ...core import log
from ...matrices import DatabaseMatrix
from ...pfam import count_pfams
from ... import constants, core


def main() -> None:
    """
    Entry point of the terminal user interface. This is called by `src/cli.py`.
    """
    try:
        print("Welcome to CoCoPyE v" + pkg_resources.get_distribution('CoCoPyE').version + ".\n")
    except pkg_resources.DistributionNotFound:
        print("Welcome to CoCoPyE.\n")

    config.init()
    print("Using configuration file at " + config.CONFIG_FILE + ".\n")

    if config.ARGS.subcommand in ["run", "database"]:
        set_num_threads(int(config.ARGS.threads))

    if config.ARGS.subcommand == "setup":
        if config.ARGS.subcommand_setup == "update-database":
            update_cocopye_db(constants.COCOPYE_DB, config.CONFIG["external"]["cocopye_db"])
            sys.exit(0)
        if config.ARGS.subcommand_setup == "cleanup":
            cleanup()
            sys.exit(0)
        if config.ARGS.subcommand_setup == "testrun":
            with tempfile.TemporaryDirectory() as tmpdir:
                command = (["cocopye"]
                           + (["--offline"] if config.ARGS.offline else [])
                           + (["--pfam24"] if config.ARGS.pfam24 else [])
                           + ["run", "-i", os.path.join(config.CONFIG["external"]["cocopye_db"], "testdata"),
                              "-o", os.path.join(tmpdir, "tempfile.csv")])

                print("Starting testrun.\n\033[37m(" + " ".join(command) + ")\033[0m\n")
                print("=======================================================================\n")
                cocopye_process = subprocess.Popen(command)
                cocopye_process.wait()
                print("\n=======================================================================")

                if cocopye_process.returncode == 0:
                    print("\n\033[92mTestrun sucessful.\033[0m")
                else:
                    print("\n\033[91mTestrun failed.\033[0m")

            sys.exit(0)
        if config.ARGS.subcommand_setup == "download-dependencies":
            check_and_download_dependencies()
            sys.exit(0)
        else:
            print("Unrecognized subcommand. See 'cocopye setup -h' for a list of available subcommands.")
            sys.exit(1)

    check_and_download_dependencies(check_only=True)

    if config.ARGS.subcommand == "database":
        create_database()

    if config.ARGS.subcommand == "run":
        if not os.path.isdir(config.ARGS.infolder):
            print("Error: The specified input folder does not exist. Exiting.")
            sys.exit(1)

        run()

    if config.ARGS.subcommand == "web":
        web()

    if config.ARGS.subcommand is None:
        print("No subcommand selected. If you don't know which subcommand to use, you will most likely")
        print("want to use \"run\". Execute \"cocopye -h\" or \"cocopye run -h\" for usage instructions.")
        sys.exit(1)


def web():
    # This is not optimal, but currently there seems to be no way to check if an extra was selected during
    # package installation
    for dep in ["fastapi", "celery", "uvicorn"]:
        if importlib.util.find_spec(dep) is None:
            print("Please run 'pip install cocopye[web] for webserver support.")
            sys.exit(1)

    from ..web.server import run_server
    run_server()


def create_database() -> None:
    count_mat, seq_list, _ = count_pfams(
        config.CONFIG["external"]["uproc_orf_bin"],
        config.CONFIG["external"]["uproc_prot_bin"],
        os.path.join(config.CONFIG["external"]["uproc_pfam_db"], "24" if config.ARGS.pfam24 else "28"),
        config.CONFIG["external"]["uproc_models"],
        config.ARGS.infolder,
        num_threads=config.ARGS.threads
    )

    # Read and sort metadata
    metadata = pd.read_csv(config.ARGS.metadata, sep=",")
    metadata = metadata.set_index("sequence").loc[seq_list].reset_index()

    # Create universal markers, one set for each superkingdom
    superkingdoms = np.unique(metadata["superkingdom"].to_numpy())
    all_markers = {}
    for superkingdom in superkingdoms:
        inds = metadata.index[metadata["superkingdom"] == superkingdom].to_list()
        submatrix = count_mat[inds]
        universal_markers = DatabaseMatrix(submatrix).universal_markers(threshold=0.95)
        all_markers[superkingdom] = universal_markers

    # Save everything to files
    os.makedirs(config.ARGS.outfolder)
    for superkingdom in all_markers:
        np.save(os.path.join(config.ARGS.outfolder, "universal_" + superkingdom + ".npy"), all_markers[superkingdom])
    np.savez_compressed(os.path.join(config.ARGS.outfolder, "count_matrix.npz"), count_mat)
    metadata.to_csv(os.path.join(config.ARGS.outfolder, "metadata.csv"), sep=",", index=False)


def run():
    results = core.core(config.CONFIG["external"]["cocopye_db"],
                        config.CONFIG["external"]["uproc_orf_bin"],
                        config.CONFIG["external"]["uproc_prot_bin"],
                        config.CONFIG["external"]["uproc_pfam_db"],
                        config.CONFIG["external"]["uproc_models"],
                        config.ARGS.infolder,
                        24 if config.ARGS.pfam24 else 28,
                        config.ARGS.file_extension,
                        config.ARGS.threads
                        )

    log("Saving results to file")
    outfile = open(config.ARGS.outfile, "w")

    if config.ARGS.verbosity == "full":
        outfile.write("bin,stage,method,1_completeness_arc,1_contamination_arc,1_completeness_bac,1_contamination_bac,"
                      "2_completeness,2_contamination,2_num_markers,3_completeness,3_contamination,"
                      "coding_density,knn_score,taxonomy,taxonomy_level,notes\n")
    elif config.ARGS.verbosity == "extended":
        outfile.write("bin,completeness,contamination,stage,method,num_markers,coding_density,knn_score,"
                      "taxonomy,taxonomy_level,notes\n")
    else:
        outfile.write("bin,completeness,contamination,method,taxonomy,taxonomy_level,notes\n")

    for result in results:
        outfile.write(result.to_csv(config.ARGS.verbosity) + "\n")

    outfile.close()


def cleanup() -> None:
    print("This will remove the following directory with all its contents:")
    print("\t" + user_data_dir("cocopye") + "\n")
    proceed = input("Proceed? [y/N] ")
    if proceed == "y" or proceed == "Y":
        shutil.rmtree(user_data_dir("cocopye"))
        print(user_data_dir("cocopye") + " removed.")
    else:
        print("Aborted.")
        sys.exit(0)
    proceed = input("\nDo you also want to remove the user configuration at " + user_config_dir("cocopye") + "? [y/N] ")
    if proceed == "y" or proceed == "Y":
        shutil.rmtree(user_config_dir("cocopye"))
        print(user_config_dir("cocopye") + " removed.")



"""
This is the API documentation of CoCoPyE.

### Submodules

- **constants**: Some constants that are required during runtime
- **core**: High-level functions that are used by `cocopye run`. If you are looking for a way to use the tool in a
Python script (without spawning a subprocess), this might be the module that you want to use.
- **histogram**: A histogram class that is used to convert a QueryMatrix into a FeatureMatrix
- **matrices**: Several matrix classes that are required during different stages of the tool
- **pfam**: Read FASTA files and use UProC to get their Pfam counts
- **ui**: Everything that is used for interaction with the user. Contains the command line interface, the webserver, the
automatic download of dependencies and configuration-related functions.
"""



# Copyright 2023 Niklas Birth
#
# This file is part of CoCoPyE.
#
# CoCoPyE is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# CoCoPyE is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with CoCoPyE. If not, see <https://www.gnu.org/licenses/>.

"""
This module contains functions for checking the applications dependencies (binaries and data) and downloading them if
needed.

Please note that these functions are part of the cli and not intended for use as a library. In particular, the functions
may output to stdout and sterr, expect input from stdin or even terminate the program.

### Submodules:

- **bin**: Functions for checking and downloading Prodigal and UProC
- **data**: Functions for checking and downloading the Pfam database and UProC models
"""
import os
import platform
import stat
import sys
from typing import Tuple, List

import requests
from tqdm import tqdm
from appdirs import user_data_dir

from .. import config
from ..config import change_config
from ... import constants


def check_and_download_dependencies(check_only: bool = False) -> None:
    """
    This function checks if all neccessary dependencies are available. If there's something missing, it will offer the
    user to download the missing files automatically. The function writes some status messages to stdout and waits for
    user input on stdin when asking if it should download the missing parts. Furthermore, it may terminate the program
    if there are corrupted dependencies or errors during download. If this function returns successfully, it is
    (hopefully)  guaranteed that all required files are present and working.
    """
    ok, missing, error = check_dependencies()

    if len(error) > 0:
        # It might be useful to provide automatic solutions for common errors (at least try to redownload the file or
        # provide a more useful error message).
        print("\nError: Some dependencies were found, but they don't seem to work. Please check them manually.\n")
        sys.exit(1)

    if len(missing) > 0:
        if check_only:
            print("\nSome dependencies are missing.")
            print("If you have already downloaded them, you can specify their path in the configuration file.")
            print("You can also use 'cocopye setup download-dependencies' to automatically download missing files.")
            sys.exit(1)
        if config.ARGS.offline:
            print("You are in offline mode. Disable it if you want to use the automatic download.")
            sys.exit(1)

        download_dependencies(missing)

        print("Downloads finished. Rechecking dependencies...\n")
        ok, missing, error = check_dependencies()
        if len(missing) > 0 or len(error) > 0:
            print("There are still errors or missing files. This should not happen and you have likely encountered a")
            print("bug in the application. Exiting.")
            sys.exit(1)

        print("Check successful.\n")


def check_dependencies() -> Tuple[List[str], List[str], List[str]]:
    """
    This function checks the status of dependencies. It writes an overview of the result to stdout. This function is
    mainly for use by `check_and_download_dependencies`.

    :return: 3-tuple (ok, missing, error) where each element is a list of dependencies that have the respective status
    Example: `(["prodigal", "uproc"], ["pfam"], ["model"])` means that Prodigal and UProC are present and working, the
    pfam database is missing and the model is there but corrupted.
    """
    from .bin import check_uproc
    from .data import check_pfam_db, check_model, check_cocopye_db

    checks = [
        check_uproc(config.CONFIG["external"]["uproc_prot_bin"]),
        check_pfam_db(config.CONFIG["external"]["uproc_pfam_db"], "24" if config.ARGS.pfam24 else "28"),
        check_model(config.CONFIG["external"]["uproc_models"]),
        check_cocopye_db(config.CONFIG["external"]["cocopye_db"], config.ARGS.offline)
    ]

    status: Tuple[List[str], List[str], List[str]] = ([], [], [])  # ok, missing, error

    print("External dependencies:")
    for lst, identifier, output in checks:
        print(output)
        status[lst].append(identifier)
    print()

    if "outdated" in checks[3][2]:
        print("New CoCoPyE database release available.")
        print("You can update the database by running \"cocopye setup update-database\".\n")

    return status[0], status[1], status[2]


def download_dependencies(missing: List[str]) -> None:
    """
    This tries to download and install missing dependencies. During the process it prints a bunch of status and progress
    messages to stdout. If the detected operating system is not Windows or Linux it will terminate the application.
    This function is mainly for use by `check_and_download_dependencies`.

    :param missing: A list of missing dependencies as generated by `check_dependencies`
    """
    from .bin import build_uproc_prot, download_uproc_win
    from .data import download_pfam_db, download_model, download_cocopye_db

    if "uproc" in missing:
        opsys = platform.system()
        if opsys == "Linux":
            build_uproc_prot(
                constants.UPROC["SRC"],
                os.path.join(user_data_dir("cocopye"), "uproc"),
                config.ARGS.verbose
            )

            change_config(
                "external", "uproc_prot_bin",
                os.path.join(user_data_dir("cocopye"), "uproc", "bin", "uproc-prot")
            )

            change_config(
                "external", "uproc_import_bin",
                os.path.join(user_data_dir("cocopye"), "uproc", "bin", "uproc-import")
            )

            change_config(
                "external", "uproc_orf_bin",
                os.path.join(user_data_dir("cocopye"), "uproc", "bin", "uproc-orf")
            )
        elif opsys == "Windows":
            download_uproc_win(constants.UPROC["WIN"], os.path.join(user_data_dir("cocopye"), "uproc"))

            change_config(
                "external", "uproc_prot_bin",
                os.path.join(user_data_dir("cocopye"), "uproc", "uproc-prot.exe")
            )

            change_config(
                "external", "uproc_import_bin",
                os.path.join(user_data_dir("cocopye"), "uproc", "uproc-import.exe")
            )

            change_config(
                "external", "uproc_orf_bin",
                os.path.join(user_data_dir("cocopye"), "uproc", "uproc-orf.exe")
            )
        else:
            print("Automatic installation of UProC is currently only supported on Windows and Linux.")
            print("See http://uproc.gobics.de for more information on how to install UProC on your system.")
            print("Exiting.\n")
            sys.exit(1)

    if "pfam" in missing:
        download_pfam_db(
            constants.PFAM_DB,
            config.CONFIG["external"]["uproc_import_bin"],
            24 if config.ARGS.pfam24 else 28,
            config.ARGS.verbose
        )

        change_config(
            "external", "uproc_pfam_db",
            os.path.join(user_data_dir("cocopye"), "pfam_db")
        )

    if "model" in missing:
        download_model(constants.UPROC_MODEL)

        change_config(
            "external", "uproc_models",
            os.path.join(user_data_dir("cocopye"), "model")
        )

    if "cocopye_db" in missing:
        download_cocopye_db(constants.COCOPYE_DB)

        change_config(
            "external", "cocopye_db",
            os.path.join(user_data_dir("cocopye"), "cocopye_db")
        )


def download(url: str, dirname: str, fname: str, label: str, chunk_size: int = 1024) -> None:
    """
    An auxiliary function to download a file. It shows a progress bar which gets removed once the download is complete.

    :param url: Download URL
    :param dirname: Destination directory (without filename)
    :param fname: Destination file (without directory path)
    :param label: Progress bar label
    :param chunk_size: Download chunk size
    """
    os.makedirs(dirname, exist_ok=True)

    # Adapted from https://stackoverflow.com/questions/37573483/progress-bar-while-download-file-over-http-with- \
    # requests/62113263#62113263
    # Author: Yan QiDong
    # License: CC BY-SA 4.0
    resp = requests.get(url, stream=True)
    total = int(resp.headers.get('content-length', 0))
    with open(os.path.join(dirname, fname), 'wb') as file, tqdm(
        desc=label,
        total=total,
        unit='iB',
        unit_scale=True,
        unit_divisor=1024,
        ncols=100,
        leave=False
    ) as bar:
        for content in resp.iter_content(chunk_size=chunk_size):
            size = file.write(content)
            bar.update(size)

    # This is probably only necessary for Prodigal, but I assume it won't hurt in other cases
    os.chmod(os.path.join(dirname, fname), stat.S_IRUSR ^ stat.S_IWUSR ^ stat.S_IXUSR)


def _green(s: str) -> str:
    return "\033[92m" + s + "\033[0m"


def _yellow(s: str) -> str:
    return "\033[93m" + s + "\033[0m"


def _red(s: str) -> str:
    return "\033[91m" + s + "\033[0m"


_TICK = _green("✓")
_CROSS = _red("✗")



# Copyright 2023 Niklas Birth
#
# This file is part of CoCoPyE.
#
# CoCoPyE is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# CoCoPyE is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with CoCoPyE. If not, see <https://www.gnu.org/licenses/>.

"""
This module contains several matrix classes that are relevant for our work. They are based on numpy arrays.

**Matrix** is a generic class and the ancestor of the other matrix classes. You probably don't need to use it directly.

**DatabaseMatrix** and **QueryMatrix** are quite similar in the sense that they both represent Pfam count
matrices. However, for better differentiation I decided to split them up into two classes.

A **DatabaseMatrix**, as the name implies, contains the count values of our database sequences, while a
**QueryMatrix** contains the counts of the input bins (the one we want to determine completeness and contamination for).

A DatabaseMatrix can be transformed into a **FeatureMatrix** to get completeness and contamination estimates from the
neural network.
"""

from __future__ import annotations

import pickle
from datetime import datetime
from typing import TypeVar, Generic, cast, Tuple, Optional, List
import numpy as np
import numpy.typing as npt
from numba_progress import ProgressBar
import pandas as pd

from ..histogram import Histogram
from ._numba_functions import nearest_neighbors_idx_njit, nearest_neighbors_idx_njit_mat, estimates_njit, estimate_njit

T = TypeVar("T")


class Matrix(Generic[T]):
    """
    A generic class for a 2-dimensional matrix. It is intended as an ancestor for the other matrix classes in this
    module.

    Please note that even though we are using a completely generic parameter, the class must be used with numpy arrays.
    """
    _mat: T

    def __init__(self, mat: T):
        """
        :param mat: Some kind of two-dimensional numpy matrix
        """
        cast(npt.NDArray[np.generic], mat)
        assert mat.ndim == 2, "Matrix has to be 2-dimensional"  # type: ignore

        self._mat = mat

    def mat(self) -> T:
        """
        This returns the underlying numpy array in case direct access to it is required.
        """
        return self._mat

    def save_to_file(self, filename: str) -> None:
        """
        Save the matrix to a file. If the filename has a .npy extension it is saved in binary format. If the extension
        is .npz it uses some form of compression. Otherwise the matrix is stored as csv.

        This is just a convenience function. If it doesn't suit your needs, just use np.save or np.savetxt.
        """
        file_format = filename.split(".")[-1]

        if file_format == "npy":
            np.save(filename, self._mat)  # type: ignore
        elif file_format == "npz":
            np.savez_compressed(filename, self._mat)
        else:
            np.savetxt(filename, self._mat, delimiter=",")  # type: ignore

    def __str__(self) -> str:
        return str(self._mat)


class DatabaseMatrix(Matrix[npt.NDArray[np.uint8]]):
    """
    A two-dimensional matrix containing Pfam counts. It can optionally contain metadata about the underlying sequences.
    """
    _metadata: Optional[pd.DataFrame] = None

    def __init__(self, mat: npt.NDArray[np.uint8], metadata: Optional[pd.DataFrame] = None):
        """
        :param mat: 2-dimensional numpy matrix. Each row contains the counts of a reference sequences while each column
        represents a Pfam.
        :param metadata: A Pandas dataframe containing metadata for each reference sequence. It is expected that the
        same row indices of the matrix and metadata represent the same sequence.
        """
        super().__init__(mat)
        self._metadata = metadata

    def metadata(self) -> Optional[pd.DataFrame]:
        """
        :return: The dataframe with metadata (in case you need direct access to it). This will return None if the
        DatabaseMatrix contains no metadata.
        """
        return self._metadata

    def nearest_neighbors(self, vec: npt.NDArray[np.uint8], k: int) -> DatabaseMatrix:
        """
        Returns the k nearest neighbors of a vector in the database matrix.
        :param vec: The count vector you want to find neighbors for
        :param k: Number of nearest neighbors to return
        :return: Database matrix where each row is one of the nearest neighbors.
        """
        return DatabaseMatrix(self._mat[self.nearest_neighbors_idx(vec, k), :])

    def nearest_neighbors_idx(self, vec: npt.NDArray[np.uint8], k: int) -> npt.NDArray[np.int64]:
        """
        Returns the row indices of the k nearest neighbors of a vector in the databse matrix. This is mainly used by the
        `nearest_neighbors`  function, but may also be useful in other situations where one needs only the indices and
        not the actual neighbors.

        :param vec: The count vector you want to find neighbors for
        :param k: Number of nearest neighbors to return

        :return: A numpy array containing the indices of the nearest neighbors.
        """
        return nearest_neighbors_idx_njit(self._mat, vec, k)[0]

    def universal_markers(self, threshold: float = 0.95) -> npt.NDArray[np.uint32]:
        """
        Calculate (possible) universal markers based on the given count matrix.

        :param threshold: Fraction of reference sequences where a marker has to occur exatly once in order to be
        considered universal

        :return: The (column) indices of possible universal markers
        """
        return np.where(np.count_nonzero(self._mat == 1, axis=0) / self._mat.shape[0] >= threshold)[0]

    def estimate(
            self,
            vec: npt.NDArray[np.uint8],
            k: int,
            frac_eq: float = 1.0,
            knn_inds: Optional[npt.NDArray[np.uint64]] = None
    ) -> Tuple[float, float, int]:
        """
        Calculate an estimate for completeness and contamination for a vector based on common markers in the k nearest
        neighbors.
        :param vec: Input vector
        :param k: Number of nearest neighbors to consider. This parameter is ignored if knn_inds is provided.
        :param frac_eq: Fraction of similar counts within the nearest neighbors required to consider a Pfam as a
        marker
        :param knn_inds: If provided, this is used as the nearest neighbors (expects a 1D numpy array with database
        indices). If this is None, the function will determine the nearest neighbors itself.
        :return: A 3-tuple: First element is the completeness estimate, the second is the contamiation estimate
        (both between 0 and 1) and the third one is the number of markers that were used. (The last value is mainly
        intended for evaluation purposes.)
        """
        return estimate_njit(self.mat(), vec, k, frac_eq, knn_inds)


class QueryMatrix(Matrix[npt.NDArray[np.uint8]]):
    """
    A two-dimensional matrix containing Pfam counts. Contrary to the DatabaseMatrix this is intended to represent a set
    of CoCoPyE input bins/sequences.
    """
    _db_mat: Optional[npt.NDArray[np.uint8]] = None
    _db_metadata: Optional[pd.DataFrame] = None
    _k: Optional[int] = None
    _knn_inds: Optional[npt.NDArray[np.uint64]] = None
    _knn_scores: Optional[npt.NDArray[np.float32]] = None

    def __init__(self, mat: npt.NDArray[np.uint8]):
        """
        :param mat: 2-dimensional numpy matrix. Each row contains the counts of an input bin/sequence while each column
        represents a Pfam.
        """
        super().__init__(mat)

    def with_database(self, db: DatabaseMatrix, k: Optional[int] = None) -> QueryMatrix:
        """
        Add a DatabaseMatrix. This is required for almost all other functions of this class.

        :param db: DatabaseMatrix to be added.
        :param k: If provided, the function calculates and stores the k nearest neighbors for each query sequence. This
        is required for functions like estimates or taxonomy.
        """
        self._db_mat = db.mat()
        self._db_metadata = db.metadata()

        if k is not None:
            self._k = k
            self._knn_inds, self._knn_scores = nearest_neighbors_idx_njit_mat(self._db_mat, self._mat, self._k)

        return self

    def knn(self) -> Optional[Tuple[int, npt.NDArray[np.uint64]]]:
        """
        Return the previously calculated nearest neighbors. This requires a database matrix with k not None.

        :return: A Tuple containing the number of nearest neighbors as well as their indices. Returns None if this
        QueryMatrix doesn't contain any information about nearest neighbors.
        """
        if self._k is None:
            return None

        return self._k, self._knn_inds

    def knn_scores(self) -> Optional[npt.NDArray[np.float32]]:
        """
        This requires a database matrix with k not None.

        :return: Similarity scores for each query bin to its nearest neighbors (value between 0 and 1; a higher value
        indicates a higher similarity). Each row represents a bin, each column a neighbor.
        """
        if self._k is None:
            return None

        return self._knn_scores

    def preestimates(self, markers: npt.NDArray[np.uint32]) -> npt.NDArray[np.float32]:
        """
        Calculate preestimates based on universal markers.

        :param markers: An array containing (column) indices that will be used as markers

        :return: A two-dimensional numpy array. Each row contains two values which are (in this order) the completeness
        and the contamination estimate for the respective query bin.
        """
        submatrix = self._mat[:, markers]

        completeness = np.sum(np.clip(submatrix, 0, 1), axis=1) / markers.shape[0]
        contamination = np.sum(submatrix - np.clip(submatrix, 0, 1), axis=1) / markers.shape[0]

        return np.array([completeness, contamination], dtype=np.float32).T

    def estimates(
            self,
            frac_eq: float = 1.0,
            print_progress: bool = True
    ) -> Optional[npt.NDArray[np.float32]]:
        """
        Calculate a completeness and contamination estimate for all rows in the QueryMatrix based on common markers in
        the k nearest neighbors.

        This requires a database matrix with k not None.

        :param frac_eq: Fraction of similar counts within the nearest neighbors required to consider a Pfam as a marker
        :param print_progress: Print a progress bar to stdout.

        :return: A 2-dimensional numpy array. For each row in the QueryMatrix there is a row with two floats, where the
        first element ist the completeness and the second one the contamination estimate.
        """
        if self._db_mat is None:
            return None

        with ProgressBar(
                total=self.mat().shape[0],
                ncols=0,
                dynamic_ncols=False,
                desc="\033[0;37m[" + str(datetime.now()) + "]\033[0m Calculating estimates",
                disable=not print_progress
        ) as progress_bar:
            result = estimates_njit(self.mat(), self._db_mat, self._k, frac_eq, progress_bar, self._knn_inds)
        return result

    def taxonomy(self) -> Optional[List[Tuple[str, str]]]:
        """
        Use a consensus of the nearest neighbors for a taxonomy estimate. This will return the most specific taxonomic
        rank that is the same for all neighbors.

        :return: A list of strings representing taxonomy estimates, one for each query bin.
        """
        if self._db_metadata is None:
            return None

        metadata = self._db_metadata
        results = []

        for knn in self._knn_inds:
            found = False
            knn_meta = metadata.iloc[knn]
            for col in ["species", "genus", "family", "order", "class", "phylum", "superkingdom"]:
                if knn_meta[col].isna().any():
                    continue
                if np.unique(knn_meta[col].to_numpy()).shape[0] == 1:
                    results.append((knn_meta.iloc[0][col], col))
                    found = True
                    break
            if not found:
                results.append(("-", "-"))

        return results

    def into_feature_mat(
            self,
            estimates: npt.NDArray[np.float32],
            resolution: int = 10
    ) -> Optional[FeatureMatrix]:
        if self._knn_inds is None:
            return None

        hist = Histogram(resolution)

        feature_vecs = []
        for idx, row in enumerate(self._mat):
            knn_mat = self._db_mat[self._knn_inds[idx]]

            knn_hists = []
            for neighbor in knn_mat:
                knn_hists.append(hist.calc_bins_for_two_counts(row, neighbor))

            mean_hist = np.mean(np.array(knn_hists), axis=0)
            mean_hist = mean_hist / np.sum(mean_hist)

            feature_vecs.append(np.concatenate([mean_hist, estimates[idx, :2]]))

        return FeatureMatrix(np.array(feature_vecs))


class FeatureMatrix(Matrix[npt.NDArray[np.double]]):
    """
    This class contains the features that are extracted from a QueryMatrix. They can be used to get completeness and
    contamination estimates of some machine learning model (in our case most likely a neural network).
    """
    def __init__(self, mat: npt.NDArray[np.double]):
        """
        You probably don't need to call this constructor directly. Instead you might want to call
        QueryMatrix.into_feature_mat.

        :param mat: Numpy matrix containing features
        """
        super().__init__(mat)

    def ml_estimates(self, model_file: str) -> npt.NDArray[np.float32]:
        """
        Calculate completeness or contamination estimates based on some machine learning model. (If you want both
        completeness and contamination, you have to call this function twice.)

        :param model_file: An sklearn .pickle file contaning the model that should be used.

        :return: An array contaning the predictions, one value for each query bin.
        """
        mlp = pickle.load(open(model_file, "rb"))
        return mlp.predict(self._mat)


def load_u8mat_from_file(filename: str) -> npt.NDArray[np.uint8]:
    """
    This is just a convenience function to load a numpy matrix from a file. If it doesn't suit your requirements, just
    use numpy.load or numpy.loadtxt directly.

    :param filename: Filename of the matrix file. If the extension is .npy it is assumed that the content is in binary
    format. If it is .npz the file will be treated as compressed binary format contaning exactly one matrix. Otherwise
    the file will be read as csv.

    :return: The loaded matrix as a numpy array
    """
    file_format = filename.split(".")[-1]

    mat: npt.NDArray[np.uint8]
    if file_format == "npy":
        mat = np.load(filename)
    elif file_format == "npz":
        mat = np.load(filename)["arr_0"]
    else:
        mat = np.loadtxt(filename, delimiter=",", dtype=np.uint8)

    return mat



# Copyright 2023 Niklas Birth
#
# This file is part of CoCoPyE.
#
# CoCoPyE is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# CoCoPyE is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with CoCoPyE. If not, see <https://www.gnu.org/licenses/>.

import os
import sys

from cocopye.ui import terminal


def main() -> None:
    try:
        terminal.main()
    except KeyboardInterrupt:
        print("\n\nInterrupted.")
        try:
            sys.exit(130)
        except SystemExit:
            os._exit(130)


if __name__ == "__main__":
    main()



# Copyright 2023 Niklas Birth
#
# This file is part of CoCoPyE.
#
# CoCoPyE is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# CoCoPyE is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with CoCoPyE. If not, see <https://www.gnu.org/licenses/>.

import asyncio
import uuid
from datetime import datetime
from importlib import resources
import os
import subprocess
import werkzeug

from Bio import SeqIO
from fastapi import FastAPI, UploadFile, Request, WebSocket
from fastapi.staticfiles import StaticFiles
from fastapi.responses import HTMLResponse
from jinja2 import Environment, PackageLoader, select_autoescape
import uvicorn
import shutil

from .. import config
from .tasks import estimate_task
from ..config import parse_config

CONFIG = parse_config()[1]

app = FastAPI()
jinja_env = Environment(
    loader=PackageLoader("cocopye.ui.web"),
    autoescape=select_autoescape()
)


@app.get("/", response_class=HTMLResponse)
async def root(request: Request):
    return jinja_env.get_template("index.html").render(
        request=request,
        contact=CONFIG["server"]["site_variables"]["contact"],
        limit=CONFIG["server"]["site_variables"]["upload_limit"],
        imprint=CONFIG["server"]["site_variables"]["imprint"],
        privacy=CONFIG["server"]["site_variables"]["privacy_policy"]
    )


@app.post("/upload")
async def upload(file: UploadFile):
    ws_id = str(uuid.uuid4())
    tmpdir = CONFIG["server"]["tmpdir"]

    os.makedirs(os.path.join(tmpdir, ws_id))

    with open(os.path.join(tmpdir, ws_id, werkzeug.utils.secure_filename(file.filename) + ".fna"), "wb") as outfile:
        # Not sure about this one. Maybe we should try to use some kind of non-blocking copy.
        shutil.copyfileobj(file.file, outfile)

    return {"ws_id": ws_id}


@app.websocket("/ws/{client_id}")
async def ws_endpoint(ws: WebSocket, client_id: str):
    await ws.accept()

    infolder = os.path.join(CONFIG["server"]["tmpdir"], client_id)
    infile = os.listdir(infolder)[0]

    await ws.send_json({"status": "progress", "content": "Checking input file"})
    if invalid_fasta(os.path.join(infolder, infile)):
        await ws.send_json({"status": "error", "content": "Invalid input file."})
        await ws.close()
        shutil.rmtree(infolder)
        return

    log("Starting new task")
    task = estimate_task.delay(
        CONFIG,
        24 if os.environ["COCOPYE_PFAM24"] == "1" else 28,
        infolder,
        CONFIG["server"]["debug"]
    )

    if task.state == "PENDING":
        await ws.send_json({"status": "progress", "content": "Waiting for task execution"})

    while task.state == "PENDING":
        # This is the best solution I found. There does not seem to be somthing like await estimates and estimates.get()
        # is a blocking call which freezes the webserver until the task is completed. (I'm actually not sure why this is
        # the case. I would have thought that it should be possible to work on three requests simultaneously if we start
        # uvicorn with three workers, but alredy the second one freezes.)
        await asyncio.sleep(0.3)

    await ws.send_json({"status": "progress", "content": "Running CoCoPyE"})

    while task.state == "RUNNING":
        await asyncio.sleep(0.3)

    if task.state == "SUCCESS":
        result = task.get()
        await ws.send_json({"status": "result", "content": result})
        log("Successfully completed task")
    else:
        await ws.send_json({"status": "error", "content": "Something went wrong."})
        log("Failed task")

    await ws.close()
    shutil.rmtree(infolder)


app.mount(
    "/",
    StaticFiles(directory=str(resources.files("cocopye.ui.web.static").joinpath("")), html=True), name="static"
)


def log(message: str):
    f = open(os.path.join(CONFIG["server"]["logdir"], "server.log"), "a")
    f.write("[" + str(datetime.now()) + "] " + message + "\n")
    f.close()


def invalid_fasta(file_path):
    empty = True
    invalid = False

    try:
        for record in SeqIO.parse(file_path, "fasta"):
            empty = False
            if any(c not in "ACGTNacgtn" for c in record.seq):
                invalid = True
    except Exception:
        invalid = True

    return empty or invalid


def run_server():
    os.environ["COCOPYE_PFAM24"] = "1" if config.ARGS.pfam24 else "0"
    os.makedirs(CONFIG["server"]["logdir"], exist_ok=True)

    # It seems to be quite difficult to get configuration options into the tasks module when it is called by celery.
    # Environment variables are probably not the best solution, but at least they work.
    celery_env = os.environ.copy()
    # ...except when they don't. But this is something we can fix later as it is not that important.
    # celery_env["CELERY_BROKER_URL"] = CONFIG["server"]["celery"]["broker"]
    # celery_env["CELERY_RESULT_BACKEND"] = CONFIG["server"]["celery"]["backend"]
    celery_env["CELERY_TIME_LIMIT"] = str(config.CONFIG["server"]["celery"]["time_limit"])
    if celery_env["CELERY_TIME_LIMIT"] == "0":
        celery_env["CELERY_TIME_LIMIT"] = "10000000"

    print("Running webserver on http://" + config.CONFIG["server"]["host"] + ":" + str(config.CONFIG["server"]["port"]))

    if config.CONFIG["server"]["debug"]:
        subprocess.Popen(
            ["celery", "-A", "cocopye.ui.web.tasks", "worker", "-c", str(config.CONFIG["server"]["celery"]["workers"])],
            env=celery_env
        )

        uvicorn.run(
            "cocopye.ui.web.server:app",
            host=config.CONFIG["server"]["host"],
            port=config.CONFIG["server"]["port"],
            workers=config.CONFIG["server"]["workers"],
        )
    else:
        subprocess.Popen(
            ["celery", "-A", "cocopye.ui.web.tasks", "worker", "-c", str(config.CONFIG["server"]["celery"]["workers"]),
                "--without-heartbeat", "--without-gossip", "--without-mingle"], env=celery_env, stdout=subprocess.DEVNULL
        )

        uvicorn.run(
            "cocopye.ui.web.server:app",
            host=config.CONFIG["server"]["host"],
            port=config.CONFIG["server"]["port"],
            workers=config.CONFIG["server"]["workers"],
            log_level="error"
        )
